@@ -402,7 +402,13 @@ static char *path_expand(struct tree *tree, struct tree *start,
     return path;
 }
 
-static struct segment *aug_tree_create(struct path *path) {
+/* Expand the tree ROOT so that it contains all components of PATH. PATH
+ * must have been initialized against ROOT by a call to PATH_FIND_ONE.
+ *
+ * Return the first segment that was created by this operation, or NULL on
+ * error.
+ */
+static struct segment *tree_create(struct tree *root, struct path *path) {
     struct segment *s, *seg;
 
     for (seg = path->segments;
@@ -420,8 +426,13 @@ static struct segment *aug_tree_create(struct path *path) {
         s->tree->children = (s+1)->tree;
     }
 
-    return seg;
+    struct tree *tree = seg_parent(path, seg);
+    if (tree == NULL)
+        list_append(root, seg->tree);
+    else
+        list_append(tree->children, seg->tree);
 
+    return seg;
  error:
     for (s = seg; s->tree != NULL && s <= last_segment(path); s++) {
         free_tree(s->tree);
@@ -574,14 +585,8 @@ struct tree *tree_set(struct tree *root, const char *path, const char *value) {
         goto error;
 
     if (r == 0) {
-        struct segment *seg = aug_tree_create(p);
-        if (seg == NULL)
+        if (tree_create(root, p) == NULL)
             goto error;
-        tree = seg_parent(p, seg);
-        if (tree == NULL)
-            list_append(root, seg->tree);
-        else
-            list_append(tree->children, seg->tree);
     }
     tree = last_segment(p)->tree;
     free_path(p);
@@ -766,6 +771,64 @@ int aug_tree_replace(struct augeas *aug, const char *path, struct tree *sub) {
     return -1;
 }
 
+int aug_mv(struct augeas *aug, const char *src, const char *dst) {
+    struct tree *root = aug->tree;
+    struct path *s = make_path(root, src);
+    struct path *d = make_path(root, dst);
+    struct tree *ts, *td, *t;
+    int r, ret;
+
+    ret = -1;
+    if (s == NULL || d == NULL)
+        goto done;
+
+    r = path_find_one(s);
+    if (r != 1)
+        goto done;
+
+    r = path_find_one(d);
+    if (r == -1)
+        goto done;
+
+    if (r == 0) {
+        if (tree_create(root, d) == NULL)
+            goto done;
+    }
+
+    ts = last_segment(s)->tree;
+    for_each_segment(ds, d) {
+        /* Don't move SRC into its own descendent */
+        if (ds->tree == ts)
+            goto done;
+    }
+    td = last_segment(d)->tree;
+    free_tree(td->children);
+
+    td->children = ts->children;
+
+    free(td->value);
+    td->value = ts->value;
+
+    ts->value = NULL;
+    ts->children = NULL;
+
+
+    t = seg_parent(s, last_segment(s));
+    if (t == NULL) t = root;
+    list_remove(ts, t->children);
+
+    t->dirty = 1;
+    td->dirty = 1;
+
+    free_tree(ts);
+
+    ret = 0;
+ done:
+    free_path(s);
+    free_path(d);
+    return ret;
+}
+
 int aug_match(const struct augeas *aug, const char *pathin, char ***matches) {
     struct path *p = NULL;
     struct tree *tree;
