@@ -22,6 +22,8 @@
  *
  */
 
+#include <config.h>
+
 #include <stdlib.h>
 #include <stddef.h>
 #include <assert.h>
@@ -964,6 +966,7 @@ int main(void)
 		    puts("out of memory");
 		    free((void *) key);
 		    free(val);
+                    break;
 		}
 
 		if (!hash_alloc_insert(h, key, val)) {
