@@ -22,6 +22,8 @@
 
 #include <config.h>
 
+#include "canonicalize.h"
+
 #include <fnmatch.h>
 #include <glob.h>
 
@@ -192,7 +194,8 @@ static char *err_path(const char *filename) {
  */
 static int store_error(struct augeas *aug,
                        const char *filename, const char *path,
-                       const char *status, const struct lns_error *err) {
+                       const char *status, int errnum,
+                       const struct lns_error *err) {
     char *ep = err_path(filename);
     size_t ep_len;
     int r;
@@ -243,6 +246,14 @@ static int store_error(struct augeas *aug,
             r = aug_set(aug, ep, err->message);
             if (r < 0)
                 goto done;
+        } else if (errnum != 0) {
+            const char *msg = strerror(errnum);
+            r = pathjoin(&ep, 1, err_msg_node);
+            if (r < 0)
+                goto done;
+            r = aug_set(aug, ep, msg);
+            if (r < 0)
+                goto done;
         }
     }
 
@@ -347,7 +358,8 @@ static int load_file(struct augeas *aug, struct lens *lens, char *filename) {
 
     result = 0;
  done:
-    store_error(aug, filename + strlen(aug->root) - 1, path, err_status, err);
+    store_error(aug, filename + strlen(aug->root) - 1, path, err_status,
+                errno, err);
     free_lns_error(err);
     free(path);
     free_tree(tree);
@@ -384,6 +396,7 @@ int transform_save(struct augeas *aug, struct transform *xform,
                    const char *path, struct tree *tree) {
     FILE *fp = NULL;
     char *augnew = NULL, *augorig = NULL, *augsave = NULL;
+    char *augorig_canon = NULL;
     char *text = NULL;
     const char *filename = path + strlen(AUGEAS_FILES_TREE) + 1;
     const char *err_status = NULL;
@@ -415,8 +428,10 @@ int transform_save(struct augeas *aug, struct transform *xform,
     // to be able to write augnew, but we have no idea what permissions
     // etc. they should get. Just the process default ?
     fp = fopen(augnew, "w");
-    if (fp == NULL)
+    if (fp == NULL) {
+        err_status = "open_augnew";
         goto done;
+    }
 
     if (tree != NULL)
         lns_put(fp, xform->lens, tree->children, text, &err);
@@ -433,6 +448,16 @@ int transform_save(struct augeas *aug, struct transform *xform,
     }
 
     if (!(aug->flags & AUG_SAVE_NEWFILE)) {
+        augorig_canon = canonicalize_file_name(augorig);
+        if (augorig_canon == NULL) {
+            if (errno == ENOENT) {
+                augorig_canon = augorig;
+            } else {
+                err_status = "canon_augorig";
+                goto done;
+            }
+        }
+
         if (aug->flags & AUG_SAVE_BACKUP) {
             int r;
             r = asprintf(&augsave, "%s%s" EXT_AUGSAVE, aug->root, filename);
@@ -441,12 +466,12 @@ int transform_save(struct augeas *aug, struct transform *xform,
                 goto done;
             }
 
-            if (rename(augorig, augsave) != 0) {
+            if (rename(augorig_canon, augsave) != 0) {
                 err_status = "rename_augsave";
                 goto done;
             }
         }
-        if (rename(augnew, augorig) != 0) {
+        if (rename(augnew, augorig_canon) != 0) {
             err_status = "rename_augnew";
             goto done;
         }
@@ -454,12 +479,14 @@ int transform_save(struct augeas *aug, struct transform *xform,
     result = 0;
 
  done:
+    store_error(aug, filename, path, err_status, errno, err);
     lens_release(xform->lens);
     free(text);
     free(augnew);
+    if (augorig_canon != augorig)
+        free(augorig_canon);
     free(augorig);
     free(augsave);
-    store_error(aug, filename, path, err_status, err);
     free_lns_error(err);
 
     if (fp != NULL)
